function [ suggest best_acc best_sv ] = sa_svc ( svm, w, c, lcr, epsilon, n_fold, init_T, end_T, dt, lambda, big_gamma, also_plot )
% 
%   function [ suggest best_acc best_sv ] = sa_svc_lin ( svm, w, c, lcr, epsilon, n_fold, init_T, end_T, dt, lambda, big_gamma, also_plot )
%       (linear kernel, N-fold cross-validation, with simulated annealing rather than grid search)
% 
% 
% Copyright (C) 2006, Matthew D. Boardman and Thomas P. Trappenberg, Dalhousie University, Nova Scotia, Canada
% 
%     This program is free software; you can redistribute it and/or modify
%     it under the terms of the GNU General Public License as published by
%     the Free Software Foundation; either version 2 of the License, or
%     (at your option) any later version.
% 
%     This program is distributed in the hope that it will be useful,
%     but WITHOUT ANY WARRANTY; without even the implied warranty of
%     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%     GNU General Public License for more details.
% 
%     You should have received a copy of the GNU General Public License along
%     with this program; if not, write to the Free Software Foundation, Inc.,
%     51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
% 

    %--------------------------------------------------------------
    % set defaults
    %--------------------------------------------------------------

    if nargin < 12
        also_plot=0;        % default to no plot
    end;

    if nargin < 11
        big_gamma=1/2;      % default to square root of percentage of support vectors
    end;
    
    if nargin < 10
        lambda=1;           % default to equal weight for accuracy and num. SV
    end;
    
    if nargin < 9
        dt=0.01;            % default drop in temperature (1%)
    end;

    if nargin < 8
        end_T = 1;          % default end temperature (1)
    end;

    if nargin < 7
        init_T = 100;       % default initial temperature (100)
    end;

    if nargin < 6
        n_fold = 10;        % default to 10-fold cross-validation
    end;

    if nargin < 5
        epsilon = 0.001;    % default to epsilon=0.001 
    end;

    if nargin < 4
        lcr = [ -5.5:1.0:15.5 ];   % wide range of log(c) values
    end;

% be well-behaved in a crash:  set output variables to some value
        suggest=0;
        best_acc=0;
        best_sv=0;

    if nargin < 3

        fprintf('\nUsage:\n\n   [ suggest best_acc best_sv ] = sa_svc_lin ( svm, w, c, lcr, epsilon, n_fold, init_T, end_t, dt, lambda, big_gamma, also_plot )\n\n');

        fprintf('    Optimize cost and gamma parameters for support vector classifier with RBF kernel.\n');
        fprintf('    Uses simulated annealing, extrinsic regularization and intensity-weighted centre of mass.\n');
        fprintf('    This version uses Leave-One-Out (LOO) cross-validation, so should only be used for\n');
        fprintf('    small training sets (say size(w,1) is < 20 or so) otherwise computations will be long.\n\n');
        
        fprintf('These parameters must be specified:\n');
        fprintf('        svm  = 1 for LIBSVM or 2 for SVM-Light\n');
        fprintf('        w    = a matrix of training data (rows = training samples, columns = input dimensions)\n');
        fprintf('        c    = a vector of training labels (rows = training samples)\n');
        fprintf('\nThese parameters are optional:\n');
        fprintf('        lcr  = log cost range  (default is [ -5.5:15.5] )\n');
        fprintf('        lgr  = log gamma range (default is [-15.5: 5.5] )\n');
        fprintf('    epsilon  = SVM stopping criteria (default is 0.001)\n');
        fprintf('     init_T  = initial temperature for simulated annealing (default is 100)\n');
        fprintf('      end_T  = ending temperature for simulated annealing (default is 1)\n');
        fprintf('        dt   = cooling rate (default is 0.1:  must be between 0 and 1!)\n');
        fprintf('     lambda  = weight of support vector ratio in score (default is 1 = equal weight)\n');
        fprintf('   big_gamma = non-linearity for support vector ratio (default is .5 = sqrt, 1 is linear, 2 is square etc.)\n');
        fprintf('   also_plot = 1 to plot the simulated annealing path as you go (default is no)\n\n');

        fprintf('Copyright (C) 2006, Matthew D. Boardman and Thomas P. Trappenberg, Dalhousie University, Nova Scotia, Canada.\n');
        fprintf('These MATLAB scripts come with ABSOLUTELY NO WARRANTY; for details, see GNU General Public License (license.txt).\n');
        fprintf('This is free software, and you are welcome to redistribute it under certain conditions; see license.txt for details.\n');
        fprintf('\n\n\n');
        
        return;
    end;
    
    ujumps = 0.1;        % accept (small) upward jumps 10% of the time
    reset_jumps = 0.01;  % occasional reset to best positions
    

    %--------------------------------------------------------------
    % run the cross validation for simulated annealing approach
    %--------------------------------------------------------------

    % set initial point
    T = init_T;
    log_cost = 0;
    i=0;  % drop temp every 10 iterations
    j=0;  % number of accuracy points found
    
    % guess that there will be 3000 accuracy points in the path (more allocated as we go)
    acc = zeros(3000,3);   
    true_acc = zeros(3000,3);
    true_sv = zeros(3000,3);
    
    current_acc = 0;
    best_acc = -1e6;
    best_true_acc = 0;
    best_true_sv = 0;
    best_num = 1;
    
    min_lcr = min(lcr);  % so that we just calculate this once
    max_lcr = max(lcr);
    
    % so that we just calculate this once
    c_factor = (max_lcr-min_lcr)/init_T;
    
    if  also_plot > 0

        figure(1); hold all;
        title('Green = CV Accuracy, Blue = Support Vectors');
        xlabel('log_e(cost)');
        ylabel('Score');
        axis ([ min_lcr max_lcr 0 1 ]); 
    end;
    

    
    %-------------------------------MAIN LOOP -------------------------------------------
    while ( T > end_T )

        % determine a new jump point at random
        new_log_cost  = randn()*T*c_factor;
 
        % introduce a small bias towards the origin (0,0) to ease complexity of final solution
        bias_cost  = (-log_cost)*T*c_factor/10;
        
        distance = sqrt(new_log_cost^2 + bias_cost^2);
        new_log_cost = new_log_cost + log_cost + bias_cost;
        
        % check that we're within the lcr and lgr bounds: else jump randomly
        if new_log_cost > max_lcr 
            new_log_cost = rand()*(max_lcr-min_lcr)+min_lcr;  
        end;
        if new_log_cost < min_lcr
            new_log_cost = rand()*(max_lcr-min_lcr)+min_lcr;  
        end;

        % create parameter string for this run
        params = sprintf ( '-s 0 -t 0 -c %.8f -e %.8f', exp(new_log_cost), epsilon );
        s_fold = sprintf ( ' -v %d', n_fold);

        
%-----------------------BEGIN TRIAL---------------------------

       % get cross-validation accuracy
        if svm == 1     % LIBSVM
            ac = svmtrain(c,w,cat(2,params,s_fold));
        else            % SVMLight
            ac = svmlearn(w,c,cat(2,params,s_fold));
        end;

       % train model (on all training data) to get number of SV
        if svm == 1     % LIBSVM
            model = svmtrain(c,w,params);
        else            % SVMLight
            model = svmlearn(w,c,params);
        end;

        new_acc = ac;
        
        new_sv = model.totalSV;        
        %new_sv = model{3};            % if above line causes an error, use this instead (old versions etc.)
    
        fprintf('Number of support vectors = %d\n', new_sv);
        
%----------------------END TRIAL--------------------------------        

        % these may be kept for later display purposes:
        true_acc(j+1,1)=0;
        true_acc(j+1,2)=new_log_cost;
        true_acc(j+1,3)=new_acc;

        true_sv(j+1,1)=0;
        true_sv(j+1,2)=new_log_cost;
        true_sv(j+1,3)=new_sv;

        % note: lambda=1 for equal weight
        new_acc = new_acc - lambda*100*((new_sv/size(w,1))^(big_gamma));
        
        fprintf ('#%4d: (Cost=exp(%3.1f),Dist=%.3f): Accuracy=(%.2f->%.2f):Temp=%.1f:', j+1, new_log_cost, distance, current_acc, new_acc, T );

        % add to the list of known positions
        j=j+1;
        acc(j,1) = 0;      
        acc(j,2) = new_log_cost;
        acc(j,3) = new_acc; 
        
        % do we want to accept this jump?
        if (new_acc >= current_acc) | (j==1)
            % descent (or flat):  accept always
            accept = 1;
            fprintf('Good');
        else
            % ascent:  maybe accept, if it's not too big of a jump upwards
            if (abs(current_acc-new_acc) < (current_acc * 0.1) )  & ( rand() < ujumps )  % allow 10% downwards (10% of the time)
           % if  rand() < ujumps   % allow upward jumps (10% of the time)
                accept = 1;
                fprintf('Accepting bad jump!!');
            else
                accept = 0;
                fprintf('Bad');
            end;
        end;
        
        % plot the jump patterns
        if also_plot > 0

            jc = 'k';   % jumps not taken
            figure(1);
            plot(new_log_cost,true_acc(j,3)/100,'gx','MarkerSize',4,'LineWidth',1);
            plot(new_log_cost,new_sv/size(w,1),'bx','MarkerSize',4,'LineWidth',1);
       
%             if accept > 0 
%                 plot([ log_cost new_log_cost ],[ log_gamma new_log_gamma ],jc);
%             end;

            drawnow;
        end;
        
        % perform jump if accepted
        if accept > 0
            current_acc = new_acc;
            log_cost = new_log_cost;    
        end;

       % check best accuracy found so far
        if current_acc > best_acc
            best_acc = current_acc;
            clear best_list;
            best_num = 1;
            best_list(1,1) = log_cost;
            best_list(1,2) = 0;
            best_true_acc = true_acc(j,3);
            best_true_sv = new_sv;
        elseif current_acc == best_acc
            best_num = best_num + 1;
            best_list ( best_num, 1 ) = log_cost;
            best_list ( best_num, 2 ) = 0;
        end;

        fprintf('\t\t(Best:%.2f=%.2f+%.2f,lC=%.4f)\n', best_acc, best_true_acc, best_true_sv, best_list(1,1) );
        
        % drop the temperature by dt every 10 iterations 
        % (regardless of jump acceptance!)
        i=i+1;
        if i >= 10
            fprintf('*** Dropping temperature from %.1f ', T);
            T = T*(1-dt);     
            i = 0;
            fprintf('to %.1f\n', T);
        end;
        
        % maybe (but rarely!) reset to a random spot on the best_list,
        %   IFF current accuracy is less than 90% of best accuracy found
        if ( rand() < reset_jumps ) & ( current_acc < best_acc*.9 )

            jump_to = ceil( rand() * best_num );  % pick a random member of the best_list
            
            current_acc = best_acc;
            log_cost = best_list ( jump_to, 1 );
            
            fprintf('*** Randomly jumping back to best_list (%d of %d)\n', jump_to, best_num );
        end;
        
    end;


%-----------------------END MAIN LOOP--------------------------



%--------------------------------------------------------------
% calculate score-weighted centre of mass
%--------------------------------------------------------------

    fprintf('End of simulated annealing.\n\nCalculating suggested parameters ...\n\n');

    % get suggested point as minimum distance from (0,0)
    for i=1:best_num
        best_list(i,3) = sqrt(best_list(i,1)^2+best_list(i,2)^2);
    end;
    best_list=sortrows(best_list,3);
    suggest=best_list(1,1:2);
    
    
    % get overall best suggested point(s)
   clear best_list;
   best_num=0;
   psuggest=suggest;
   suggest=zeros(2,1);
   denom=0;

   acc=acc(1:j,:);   % cut off excess points if needed (since we guessed at the number of iterations)
   
   
   for i=1:j               
         if (sqrt((acc(i,1)-psuggest(1,2))^2+(acc(i,2)-psuggest(1,1))^2) < 1 )   % params are within log radius of 1?
            if ((best_acc > 0) & (acc(i,3) >= best_acc*.98)) | (best_acc < 0) & (acc(i,3) >= best_acc*1.02) % score is within 2% of best point?

                if best_acc > 0
                    suggest(1)=suggest(1)+(acc(i,3))*acc(i,2);  % cost
                    suggest(2)=suggest(2)+(acc(i,3))*acc(i,1);  % gamma
                    denom=denom+(acc(i,3));                     % score
                else    % reverse if scores are negative (not really ideal ... any ideas?)
                    suggest(1)=suggest(1)+(100+acc(i,3))*acc(i,2);  % cost
                    suggest(2)=suggest(2)+(100+acc(i,3))*acc(i,1);  % gamma
                    denom=denom+(100+acc(i,3));                     % score
                end;
                
                best_num=best_num+1;
                
                best_list(best_num,1)=acc(i,2);  % cost
                best_list(best_num,2)=acc(i,1);  % gamma
                best_list(best_num,3)=acc(i,3);  % score
            end;
         end;
   end;
  
   suggest=suggest/denom;
   
       % run SVM again with these parameters
    params = sprintf ( '-s 0 -t 0 -c %.8f -e %.8f', exp(suggest(1)), epsilon );
    s_fold = sprintf ( ' -v %d', n_fold);

    %---------- BEGIN TRIAL -----------

       % get cross-validation accuracy
        if svm == 1     % LIBSVM
            ac = svmtrain(c,w,cat(2,params,s_fold));
        else            % SVMLight
            ac = svmlearn(w,c,cat(2,params,s_fold));
        end;

        % train model (on all training data) to get number of SV
        if svm == 1     % LIBSVM
            model = svmtrain(c,w,params);
        else            % SVMLight
            model = svmlearn(w,c,params);
        end;

        new_acc = ac;

        new_sv = model.totalSV;        
        %new_sv = model{3};            % if above line causes an error, use this instead (old versions etc.)
    
    %---------- END TRIAL -----------

    suggest_acc = new_acc;
    suggest_sv = new_sv;
   
   % keep overall best points too, for display or comparison
    best_acc = best_true_acc;
    best_sv = best_true_sv;


%--------------------------------------------------------------
% plot the results as a 3D scatterplot
%--------------------------------------------------------------

 if also_plot > 0
    plot([ suggest(1) suggest(1) ],[ best_acc/100 best_sv/size(w,1) ],'k-');
 end;


%--------------------------------------------------------------
% detect failure, otherwise print results
%--------------------------------------------------------------

if best_acc == 0
    fprintf ( '\n\n!!! Resulting accuracy could not be measured (same over the entire range of values for cost and gamma).\n');
    return;
else
    fprintf ('\n\n--------------------\nDone! Top few points:\n');

    for i=1:best_num
        fprintf ('   Score = %.1f    Cost = exp(%3.3f) = %.8f\n', ...
            best_list(i,3), best_list(i,1), exp(best_list(i,1)) );
    end;
    
    fprintf('\nBest point found: \n');
    fprintf ('   Acc = %.1f  SV = %d/%d  Cost = exp(%3.3f) = %.8f\n', ...
        best_acc, best_sv, size(w,1), psuggest(1,1), exp(psuggest(1,1)) );
    
    fprintf('\nSuggested optimum point: (intensity-weighted centre of mass of best points)\n');
    fprintf ('   Acc = %.1f  SV = %d/%d  Cost = exp(%3.3f) = %.8f\n\n\n', ...
        suggest_acc, suggest_sv, size(w,1), suggest(1), exp(suggest(1)) );

end;

